//	ArrestExe - File Audit Utility
//	Copyright (C) 2002-2006 AnonyMouse
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

// Project Includes:
#include "Stdafx.h"
#include "ArrestExe.h"
#include "AuditThread.h"
#include "ArrestExeDlg.h"
#include "FolderScanThread.h"

// General includes:
#ifdef _MBCS
#include <atlconv.h>
#endif

// Utility includes:
#include "FileUtils.h"
#include "StringUtils.h"
#include "CommonProcUtils.h"

// Debug Codes:
#ifdef _DEBUG
#define new DEBUG_NEW
#endif

CAuditThread::CAuditThread( CString ipAuditPath, CStringArray* ipExtensionArray,bool ipCompleteDelete, bool ipUseScan, bool ipUseStartSearch, int ipPollingMilliSecond, CLogWriter* ipLogWriter ):CThreadHandler( ( ipUseScan ? ipPollingMilliSecond : 0 ), ipLogWriter )
{
	// Store Audit Options
	m_FirstRun = true;
	m_AuditPath = ipAuditPath;
	m_CompleteDelete = ipCompleteDelete;
	m_UseScan = ipUseScan;
	m_UseStartSearch = ipUseStartSearch;
	for( int i = 0; i < ipExtensionArray->GetCount(); i++ )
	{
		CString lExtension = ipExtensionArray->GetAt( i );
		lExtension.MakeLower();
		m_ExtensionArray.Add( lExtension );
	}

	// Create Thread Manager
	m_ScanThreadManager = new CThreadManager( m_LogWriter );

	// Create Buffer
	m_RDCW_BufferLength = CFileUtils::IsNetwork( m_AuditPath ) ? AUDITTHREAD_RDCW_BUFFER_NETWORK : AUDITTHREAD_RDCW_BUFFER;
	m_RDCW_Buffer = new char[ m_RDCW_BufferLength ];

	// Initialize Hnadle
	m_FFCN = NULL;
	m_DirHandle = NULL;
	m_CompletionPort = NULL;

	// Create Handle
	m_InitSuccess = false;
	if( m_UseScan )
	{
		// Create FFCN Handle
		m_FFCN = FindFirstChangeNotification( m_AuditPath, TRUE, AUDITTHREAD_FFCN_NOTIFY_FLAG );
		if( m_FFCN != INVALID_HANDLE_VALUE )
		{
			m_InitSuccess = true;
		}
	}
	else
	{
		// Create Directory Handle
		m_DirHandle = CreateFile( m_AuditPath, FILE_LIST_DIRECTORY, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS | FILE_FLAG_OVERLAPPED, NULL );
		if( m_DirHandle != INVALID_HANDLE_VALUE )
		{
			// Create I/O Completion Port
			m_CompletionKey = 0;
			m_CompletionPort = CreateIoCompletionPort( m_DirHandle, NULL, ( ULONG_PTR )( &m_CompletionKey ), 0 );
			if( m_CompletionPort != NULL )
			{
				m_InitSuccess = true;
			}
			else
			{
				CloseHandle( m_DirHandle );
			}
		}
	}
}

CAuditThread::~CAuditThread()
{
	// Close Handles
	if( m_InitSuccess )
	{
		if( m_UseScan )
		{
			FindCloseChangeNotification( m_FFCN );
		}
		else
		{
			CloseHandle( m_CompletionPort );
			CloseHandle( m_DirHandle );
		}
	}

	// Delete Buffer
	delete m_RDCW_Buffer;

	// Delete Thread Manager
	delete m_ScanThreadManager;
}

void CAuditThread::OnEndThread()
{
	// Cancel All Thread
	m_ScanThreadManager->Lock();
	m_ScanThreadManager->CancelAllThread();
	m_ScanThreadManager->Unlock();

	// Wait For Cancel Complete
	while( 1 )
	{
		if( m_ScanThreadManager->GetActiveThreadCount() == 0 )
		{
			break;
		}
		Sleep( AUDITTHREAD_SCANWAIT_POLLING_MILLISECOND );
	}
}

UINT CAuditThread::RunImpl()
{
	// Check Directory Handler Valid
	if( !m_InitSuccess )
	{
		WriteLog( CStringUtils::GetResStr( IDS_LOG_DEL_A ) + CStringUtils::GetDateStr() + CStringUtils::GetResStr( IDS_LOG_ERR_INITFAIL ) + _T( "\r\n" ) + m_AuditPath );
		return CThreadHandler::THREAD_ERROR;
	}

	// If First Run, Scan Folder
	if( m_FirstRun )
	{
		if( m_UseStartSearch )
		{
			if( !ScanFolder( m_AuditPath ) )
			{
				return CThreadHandler::THREAD_DONE;
			}
		}
		m_FirstRun = false;
	}

	// Wait For File Change
	if( m_UseScan )
	{
		// Wait For FFCN
		DWORD dReturn = WaitForSingleObject( m_FFCN, AUDITTHREAD_FFCN_POLLING_MILLISECOND );

		// Check Object Alive
		if( dReturn == WAIT_ABANDONED )
		{
			WriteLog( CStringUtils::GetResStr( IDS_LOG_DEL_A ) + CStringUtils::GetDateStr() + CStringUtils::GetResStr( IDS_LOG_ERR_OBJECT_ABANDONED ) + _T( "\r\n" ) + m_AuditPath );
			return CThreadHandler::THREAD_ERROR;
		}

		// Get Next FFCN
		FindNextChangeNotification( m_FFCN );

		// Check TimeOut
		if( dReturn == WAIT_TIMEOUT )
		{
			return CThreadHandler::THREAD_RESUME;
		}

		// Wait For Delete Complete( NoWait Scan Uses Too Much CPU Power )
		if( !WaitDeleteComplete() )
		{
			return CThreadHandler::THREAD_DONE;
		}

		// Schan Folder
		if( !ScanFolder( m_AuditPath ) )
		{
			return CThreadHandler::THREAD_DONE;
		}
	}
	else
	{
		// Check Directory Exist
		if( !CFileUtils::IsFileExist( m_AuditPath ) )
		{
			WriteLog( CStringUtils::GetResStr( IDS_LOG_DEL_A ) + CStringUtils::GetDateStr() + CStringUtils::GetResStr( IDS_LOG_ERR_OBJECT_NOTEXIST ) + _T( "\r\n" ) + m_AuditPath );
			return CThreadHandler::THREAD_ERROR;
		}

		// Read Directory Change ( Async )
		OVERLAPPED lOverLapped; memset( &lOverLapped, 0 ,sizeof( lOverLapped ) );
		BOOL lRetRead = ReadDirectoryChangesW( m_DirHandle, m_RDCW_Buffer, m_RDCW_BufferLength, TRUE, AUDITTHREAD_RDCW_NOTIFY_FLAG, NULL, &lOverLapped, NULL );
		if( !lRetRead )
		{
			WriteLog( CStringUtils::GetResStr( IDS_LOG_DEL_A ) + CStringUtils::GetDateStr() + CStringUtils::GetResStr( IDS_LOG_ERR_READDIR_FAILED ) + _T( "\r\n" ) + m_AuditPath );
			return CThreadHandler::THREAD_ERROR;
		}

		// Get Completion Status ( Wait For Complete )
		LPOVERLAPPED lOverLappedPtr = &lOverLapped;
		DWORD lByteReturned;
		BOOL lRetGetStatus = GetQueuedCompletionStatus( m_CompletionPort, &lByteReturned, &m_CompletionKey, &lOverLappedPtr, INFINITE );

		// Check Error
		if( !lRetGetStatus )
		{
			WriteLog( CStringUtils::GetResStr( IDS_LOG_DEL_A ) + CStringUtils::GetDateStr() + CStringUtils::GetResStr( IDS_LOG_ERR_READDIR_FAILED ) + _T( "\r\n" ) + m_AuditPath );
			return CThreadHandler::THREAD_ERROR;
		}

		// Check Timeout
		if( m_CompletionKey == 0 )
		{
			return CThreadHandler::THREAD_DONE;
		}

		if( lByteReturned != 0 )
		{
			// Check All Changes
			int lPosition = 0;
			while( 1 )
			{
				// Get Notify Information
				FILE_NOTIFY_INFORMATION* lNofityInfo = ( FILE_NOTIFY_INFORMATION* )&( m_RDCW_Buffer[ lPosition ] );
				DWORD lNextEntryOffset = lNofityInfo->NextEntryOffset;
				DWORD lAction = lNofityInfo->Action;
				DWORD lFileNameLength = lNofityInfo->FileNameLength;
				wchar_t lFileName[ FILE_UTILS_MAX_PATH ]; memset( lFileName, 0, sizeof( lFileName ) );

				// Ignore Remove And Modify Notification
				if( lAction == FILE_ACTION_ADDED || lAction == FILE_ACTION_RENAMED_NEW_NAME  )
				{
					memcpy( lFileName, lNofityInfo->FileName, lFileNameLength );
#ifdef _UNICODE
					CString lFilePath = m_AuditPath + _T( "\\" ) + lFileName;
#else _MBCS
					USES_CONVERSION;
					CString lFilePath = m_AuditPath + _T( "\\" ) + W2A( lFileName );
#endif
					// If Path Is File, Request Delete. If Path Is Directory, Scan Folder.
					if( !CFileUtils::IsDirectory( lFilePath ) )
					{
						CCommonProcUtils::RequestDelete( lFilePath, &m_ExtensionArray );
					}
					else{
						// Scan, If Directory Added
						if( lAction == FILE_ACTION_ADDED )
						{
							ScanFolder( lFilePath, true );
						}
					}
				}

				// Shift To Next Position
				if( lNextEntryOffset != 0 )
					lPosition += lNextEntryOffset;
				else
					break;
			}
		}
		else
		{
			// If lByteReturned Is 0 There Are Too Much Changes( Need Manual Check )

			// Wait For Delete Complete( NoWait Scan Uses Too Much CPU Power )
			if( !WaitDeleteComplete() )
			{
				return CThreadHandler::THREAD_DONE;
			}

			// Scan Folder
			ScanFolder( m_AuditPath );
		}
	}

	return CThreadHandler::THREAD_RESUME;
}

bool CAuditThread::ScanFolder( CString ipFilePath, bool ipAsync )
{
	// Lock Thread Manager
	m_ScanThreadManager->Lock();
	
	// Create Folder Scan Thread
	CFolderScanThread* lFolderScanThread = new CFolderScanThread( ipFilePath, &m_ExtensionArray, m_CompleteDelete, 0, m_LogWriter );
	CWinThread* lWinThread = AfxBeginThread( CThreadManager::RunThread, ( LPVOID )lFolderScanThread, THREAD_PRIORITY_IDLE, 0, CREATE_SUSPENDED );
	lFolderScanThread->SetHandle( lWinThread->m_hThread );
	m_ScanThreadManager->AddThread( lFolderScanThread );
	lWinThread->ResumeThread();
	
	// Unlock Thread Manager
	m_ScanThreadManager->Unlock();

	// Sync Thread
	if( !ipAsync )
	{
		bool bCancelDone = false;
		while( 1 )
		{
			// Check Thread End
			if( m_ScanThreadManager->GetActiveThreadCount() == 0 )
			{
				break;
			}
			// Cancel Check
			if( IsCanceled() && !bCancelDone )
			{
				m_ScanThreadManager->Lock();
				m_ScanThreadManager->CancelAllThread();
				m_ScanThreadManager->Unlock();
				bCancelDone = true;
			}
			Sleep( AUDITTHREAD_SCANWAIT_POLLING_MILLISECOND );
		}
	}

	return !IsCanceled();
}

bool CAuditThread::WaitDeleteComplete()
{
	while( 1 )
	{
		if( CAppUtils::GetDeleteServiceThread()->GetQueCount() == 0 )
		{
			break;
		}
		Sleep( AUDITTHREAD_DELETE_POLLING_MILLISECOND );
		if( IsCanceled() )
		{
			return false;
		}
	}
	return true;
}

void CAuditThread::Cancel()
{
	if( !m_UseScan && m_InitSuccess )
	{
		PostQueuedCompletionStatus( m_CompletionPort, 0, 0, NULL );
	}
	CThreadHandler::Cancel();
}
